package com.grizzly.test.arp.client;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;

import org.apache.commons.httpclient.ChunkedInputStream;
import org.apache.commons.httpclient.HttpMethod;
import org.apache.commons.httpclient.util.EncodingUtil;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

public class ChunkedReader
{
	/** The inputstream that we're using */
	private InputStream in;

	/** The chunk size */
	private int chunkSize;

	/** True if we'are at the beginning of stream */
	private boolean bof = true;

	/** True if we've reached the end of stream */
	private boolean eof = false;

	/** True if this stream is closed */
	private boolean closed = false;

	/** The method that this stream came from */
	private HttpMethod method = null;

	/** Log object for this class. */
	private static final Log LOG = LogFactory.getLog(ChunkedInputStream.class);

	/**
	 * ChunkedInputStream constructor that associates the chunked input stream with a 
	 * {@link HttpMethod HTTP method}. Usually it should be the same {@link HttpMethod 
	 * HTTP method} the chunked input stream originates from. If chunked input stream 
	 * contains any footers (trailing headers), they will be added to the associated 
	 * {@link HttpMethod HTTP method}.
	 *
	 * @param in the raw input stream
	 * @param method the HTTP method to associate this input stream with. Can be <tt>null</tt>.  
	 *
	 * @throws IOException If an IO error occurs
	 */
	public ChunkedReader(final InputStream in, final HttpMethod method) throws IOException
	{
		if (in == null)
		{
			throw new IllegalArgumentException("InputStream parameter may not be null");
		}
		this.in = in;
		this.method = method;
	}

	/**
	 * ChunkedInputStream constructor
	 *
	 * @param in the raw input stream
	 *
	 * @throws IOException If an IO error occurs
	 */
	public ChunkedReader(final InputStream in) throws IOException
	{
		this(in, null);
	}

	/**
	 * Read the next chunk from the stream.
	 */
	public byte[] readChunk() throws IOException
	{
		if (closed)
		{
			throw new IOException("Attempted read from closed stream.");
		}

		if (eof)
		{
			return null;
		}

		nextChunk();
		if (eof)
		{
			return null;
		}

		// System.out.println("ChunkedReader.readChunk(in=" + in.hashCode() + ") --> chunkSize=" + chunkSize);

		byte[] bytes = new byte[chunkSize];
		int count = 0;
		while (count < chunkSize)
		{
			count += in.read(bytes, count, chunkSize - count);
		}

		return bytes;
	}

	/**
	 * Read the CRLF terminator.
	 * @throws IOException If an IO error occurs.
	 */
	private void readCRLF() throws IOException
	{
		int cr = in.read();
		int lf = in.read();
		if ((cr != '\r') || (lf != '\n'))
		{
			throw new IOException("CRLF expected at end of chunk: " + cr + "/" + lf);
		}
	}

	/**
	 * Read the next chunk.
	 * @throws IOException If an IO error occurs.
	 */
	private void nextChunk() throws IOException
	{
		if (!bof)
		{
			readCRLF();
		}
		chunkSize = getChunkSizeFromInputStream(in);
		bof = false;
		if (chunkSize == 0)
		{
			close();
		}
	}

	/**
	 * Expects the stream to start with a chunksize in hex with optional
	 * comments after a semicolon. The line must end with a CRLF: "a3; some
	 * comment\r\n" Positions the stream at the start of the next line.
	 *
	 * @param in The new input stream.
	 * @param required <tt>true<tt/> if a valid chunk must be present,
	 *                 <tt>false<tt/> otherwise.
	 * 
	 * @return the chunk size as integer
	 * 
	 * @throws IOException when the chunk size could not be parsed
	 */
	private static int getChunkSizeFromInputStream(final InputStream in) throws IOException
	{

		ByteArrayOutputStream baos = new ByteArrayOutputStream();
		// States: 0=normal, 1=\r was scanned, 2=inside quoted string, -1=end
		int state = 0;
		while (state != -1)
		{
			int b = in.read();
			if (b == -1)
			{
				throw new IOException("chunked stream ended unexpectedly");
			}
			switch (state)
			{
			case 0:
				switch (b)
				{
				case '\r':
					state = 1;
					break;
				case '\"':
					state = 2;
					/* fall through */
				default:
					baos.write(b);
				}
				break;

			case 1:
				if (b == '\n')
				{
					state = -1;
				}
				else
				{
					// this was not CRLF
					throw new IOException("Protocol violation: Unexpected" + " single newline character in chunk size");
				}
				break;

			case 2:
				switch (b)
				{
				case '\\':
					b = in.read();
					baos.write(b);
					break;
				case '\"':
					state = 0;
					/* fall through */
				default:
					baos.write(b);
				}
				break;
			default:
				throw new RuntimeException("assertion failed");
			}
		}

		//parse data
		String dataString = EncodingUtil.getAsciiString(baos.toByteArray());
		int separator = dataString.indexOf(';');
		dataString = (separator > 0) ? dataString.substring(0, separator).trim() : dataString.trim();

		int result;
		try
		{
			result = Integer.parseInt(dataString.trim(), 16);
		}
		catch (NumberFormatException e)
		{
			throw new IOException("Bad chunk size: " + dataString);
		}
		return result;
	}

	/**
	 * Upon close, this reads the remainder of the chunked message,
	 * leaving the underlying socket at a position to start reading the
	 * next response without scanning.
	 * @throws IOException If an IO problem occurs.
	 */
	public void close() throws IOException
	{
		if (!closed)
		{
			try
			{
				if (!eof)
				{
					exhaustInputStream();
				}
			}
			finally
			{
				// System.out.println("ChunkedReader.close(in=" + in.hashCode() + ")");
				eof = true;
				closed = true;
			}
		}
	}

	/**
	 * Exhaust an input stream, reading until EOF has been encountered.
	 *
	 * <p>Note that this function is intended as a non-public utility.
	 * This is a little weird, but it seemed silly to make a utility
	 * class for this one function, so instead it is just static and
	 * shared that way.</p>
	 *
	 * @param inStream The {@link InputStream} to exhaust.
	 * @throws IOException If an IO problem occurs
	 */
	void exhaustInputStream() throws IOException
	{
		// read and discard the remainder of the message
		byte buffer[] = new byte[1024];
		while (in.read(buffer) >= 0)
		{
			;
		}
	}
}
