package com.grizzly.test.arp.server.util;

import java.io.IOException;
import java.io.OutputStream;

import com.sun.grizzly.tcp.http11.GrizzlyRequest;
import com.sun.grizzly.tcp.http11.GrizzlyResponse;

/**
 * Original code comes from the BaseStreamingHTTPEndpoint class of BlaseDS 3.0.x.
 * 
 * @author eVrignaud
 */
public class StreamingHelper
{
	/** Used in chunked HTTP responses frequently so initialize it statically for general use. */
	private static final byte[] CRLF_BYTES = { (byte) 13, (byte) 10 };

	/** Used for the terminal chunk within a chunked response. */
	private static final byte ZERO_BYTE = (byte) 48;

	/** Used to signal that a chunk of data should be skipped by the client. */
	private static final byte NULL_BYTE = (byte) 0;

	/**
	 * For Flex, it's mandatory to add data in an HTTP "Transfer-Encoding: chunked" format.
	 * The low level comet layer send data in a chunked format, but the WebBrowser remove all this chunk informations.
	 * The Flex client part need to cut the byte stream into messages, so we add a second time the chunk marker format.
	 */
	public static boolean isSubChunkNeeded(GrizzlyRequest request)
	{
		String value = request.getParameter("sub_chunk");
		return (value != null && value.equals("true"));
	}

	/**
	 * Helper method to write a chunk of bytes to the output stream in an HTTP
	 * "Transfer-Encoding: chunked" format.
	 * If the bytes array is null or empty, a terminal chunk will be written to
	 * signal the end of the response.
	 * Once the chunk is written to the output stream, the stream will be flushed immediately (no buffering).
	 *
	 * @param bytes The array of bytes to write as a chunk in the response; or if null, the signal to write the final chunk to complete the response.
	 * @param os The output stream the chunk will be written to.
	 * @param response The HttpServletResponse, used to flush the chunk to the client.
	 *
	 * @throws IOException if writing the chunk to the output stream fails.
	 */
	public static void streamChunk(byte[] bytes, OutputStream output, GrizzlyResponse response) throws IOException
	{
		if ((bytes != null) && (bytes.length > 0))
		{
			byte[] chunkLength = Integer.toHexString(bytes.length).getBytes("ASCII");
			output.write(chunkLength);
			output.write(CRLF_BYTES);
			output.write(bytes);
			output.write(CRLF_BYTES);
			response.flushBuffer();
		}
		else
		// Send final 'EOF' chunk for the response.
		{
			output.write(ZERO_BYTE);
			output.write(CRLF_BYTES);
			response.flushBuffer();
		}
	}
}
